/*
 * Copyright (c) 2008 Topspin Communications.  All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * Fork Id: rc_pingpong.c 5257 2006-02-01 16:56:45Z roland
 */
/*
 * Copyright (C) 2008 United States Government as represented by the
 * Administrator of the National Aeronautics and Space Administration.
 * All Rights Reserved.
 *
 * $Id: rc_pingpong.cxx,v 1.12 2008/05/11 21:24:29 bgreen Exp $
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif /* HAVE_CONFIG_H */

#include <ibvpp.h>
#include <ibvpp_utility.h>

#include <iostream>
#include <stdexcept>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netdb.h>
#include <malloc.h>
#include <getopt.h>
#include <arpa/inet.h>
#include <time.h>

using namespace std;
using tr1::shared_ptr;

enum {
	PINGPONG_RECV_WRID = 1,
	PINGPONG_SEND_WRID = 2,
};

static int page_size;

struct pingpong_context {
    ibv::comp_channel_ptr channel;
    ibv::mr_ptr           mr;
    ibv::cq_ptr           cq;
    ibv::qp_ptr           qp;
    void                  *buf;
    int                   size;
    int                   rx_depth;
    int                   pending;
};

struct pingpong_dest {
	int lid;
	int qpn;
	int psn;
};

static void pp_connect_ctx(pingpong_context *ctx, int port, int my_psn,
                           ibv_mtu mtu, pingpong_dest *dest)
{
    ibv_qp_attr attr;
    ibv::clear(attr);
    attr.qp_state		= IBV_QPS_RTR;
    attr.path_mtu		= mtu;
    attr.dest_qp_num		= dest->qpn;
    attr.rq_psn 		= dest->psn;
    attr.max_dest_rd_atomic	= 1;
    attr.min_rnr_timer		= 12;
    attr.ah_attr.is_global	= 0;
    attr.ah_attr.dlid		= dest->lid;
    attr.ah_attr.sl		= 0;
    attr.ah_attr.src_path_bits	= 0;
    attr.ah_attr.port_num	= port;

    ctx->qp->modify(&attr,
                    ibv_qp_attr_mask(
                            IBV_QP_STATE              |
                            IBV_QP_AV                 |
                            IBV_QP_PATH_MTU           |
                            IBV_QP_DEST_QPN           |
                            IBV_QP_RQ_PSN             |
                            IBV_QP_MAX_DEST_RD_ATOMIC |
                            IBV_QP_MIN_RNR_TIMER));

    attr.qp_state 	    = IBV_QPS_RTS;
    attr.timeout 	    = 14;
    attr.retry_cnt 	    = 7;
    attr.rnr_retry 	    = 7;
    attr.sq_psn 	    = my_psn;
    attr.max_rd_atomic  = 1;
    ctx->qp->modify(&attr,
                    ibv_qp_attr_mask(
                            IBV_QP_STATE              |
                            IBV_QP_TIMEOUT            |
                            IBV_QP_RETRY_CNT          |
                            IBV_QP_RNR_RETRY          |
                            IBV_QP_SQ_PSN             |
                            IBV_QP_MAX_QP_RD_ATOMIC));
}

static shared_ptr<pingpong_dest> pp_client_exch_dest(
        const char *servername, int port,
        const pingpong_dest *my_dest)
{
    struct addrinfo *res, *t;
    struct addrinfo hints;
    memset(&hints,0,sizeof(hints));
    hints.ai_family   = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    char *service;
    char msg[sizeof "0000:000000:000000"];
    int n;
    int sockfd = -1;
    shared_ptr<pingpong_dest> rem_dest;

    asprintf(&service, "%d", port);
    n = getaddrinfo(servername, service, &hints, &res);

    if (n < 0) {
        fprintf(stderr, "%s for %s:%d\n", gai_strerror(n), servername, port);
        free(service);
        throw runtime_error("getaddrinfo()");
    }

    for (t = res; t; t = t->ai_next) {
        sockfd = socket(t->ai_family, t->ai_socktype, t->ai_protocol);
        if (sockfd >= 0) {
            if (!connect(sockfd, t->ai_addr, t->ai_addrlen))
                break;
            close(sockfd);
            sockfd = -1;
        }
    }

    freeaddrinfo(res);
    free(service);

    if (sockfd < 0) {
        fprintf(stderr, "Couldn't connect to %s:%d\n", servername, port);
        throw runtime_error("connect()");
    }

    try {
        sprintf(msg, "%04x:%06x:%06x", my_dest->lid, my_dest->qpn, my_dest->psn);
        if (write(sockfd, msg, sizeof msg) != sizeof msg) {
            fprintf(stderr, "Couldn't send local address\n");
            throw runtime_error("write()");
        }

        if (read(sockfd, msg, sizeof msg) != sizeof msg) {
            perror("client read");
            fprintf(stderr, "Couldn't read remote address\n");
            throw runtime_error("read()");
        }

        write(sockfd, "done", sizeof "done");

        rem_dest = shared_ptr<pingpong_dest>(new pingpong_dest);
    } catch (...) {
        close(sockfd);
        throw;
    }

    close(sockfd);

    sscanf(msg, "%x:%x:%x", &rem_dest->lid, &rem_dest->qpn, &rem_dest->psn);

    return rem_dest;
}

static shared_ptr<pingpong_dest> pp_server_exch_dest(
        pingpong_context *ctx,
        int ib_port, enum ibv_mtu mtu, int port,
        const pingpong_dest *my_dest)
{
    struct addrinfo *res, *t;
    struct addrinfo hints;
    memset(&hints,0,sizeof(hints));
    hints.ai_flags    = AI_PASSIVE;
    hints.ai_family   = AF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;

    char *service;
    char msg[sizeof "0000:000000:000000"];
    int n;
    int sockfd = -1, connfd;
    shared_ptr<pingpong_dest> rem_dest;

    asprintf(&service, "%d", port);
    n = getaddrinfo(NULL, service, &hints, &res);

    if (n < 0) {
        fprintf(stderr, "%s for port %d\n", gai_strerror(n), port);
        free(service);
        throw runtime_error("getaddrinfo()");
    }

    for (t = res; t; t = t->ai_next) {
        sockfd = socket(t->ai_family, t->ai_socktype, t->ai_protocol);
        if (sockfd >= 0) {
            n = 1;

            setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &n, sizeof n);

            if (!bind(sockfd, t->ai_addr, t->ai_addrlen))
                break;
            close(sockfd);
            sockfd = -1;
        }
    }

    freeaddrinfo(res);
    free(service);

    if (sockfd < 0) {
        fprintf(stderr, "Couldn't bind to port %d\n", port);
        throw runtime_error("bind()");
    }

    try {
        listen(sockfd, 1);
        connfd = accept(sockfd, NULL, 0);
        close(sockfd);
        if (connfd < 0) {
            fprintf(stderr, "accept() failed\n");
            throw runtime_error("accept()");
        }

        n = read(connfd, msg, sizeof msg);
        if (n != sizeof msg) {
            perror("server read");
            fprintf(stderr, "%d/%d: Couldn't read remote address\n", n, (int) sizeof msg);
            throw runtime_error("read()");
        }

        rem_dest = shared_ptr<pingpong_dest>(new pingpong_dest);

        sscanf(msg, "%x:%x:%x", &rem_dest->lid, &rem_dest->qpn, &rem_dest->psn);

        try {
            pp_connect_ctx(ctx, ib_port, my_dest->psn, mtu, rem_dest.get());
        } catch (std::exception &e) {
            fprintf(stderr, "Couldn't connect to remote QP: %s\n", e.what());
            throw;
        }

        sprintf(msg, "%04x:%06x:%06x", my_dest->lid, my_dest->qpn, my_dest->psn);
        if (write(connfd, msg, sizeof msg) != sizeof msg) {
            fprintf(stderr, "Couldn't send local address\n");
            throw runtime_error("write()");
        }

        read(connfd, msg, sizeof msg);
    } catch (...) {
        close(connfd);
        throw;
    }

    close(connfd);
    return rem_dest;
}

static pingpong_context *pp_init_ctx(const ibv::port_ptr &port,
                                     int size,
                                     int rx_depth,
                                     int use_event)
{
    pingpong_context *ctx;

    ctx = new pingpong_context();

    ctx->size     = size;
    ctx->rx_depth = rx_depth;

    ctx->buf = memalign(page_size, size);
    if (!ctx->buf)
        throw runtime_error("Couldn't allocate work buf.");

    memset(ctx->buf, 0, size);

    ibv::context_ptr context = port->get_context();

    if (use_event)
        ctx->channel = context->create_comp_channel();

    ibv::pd_ptr pd = context->alloc_pd();

    ctx->mr = pd->reg_mr(ctx->buf, size, IBV_ACCESS_LOCAL_WRITE);

    ctx->cq = context->create_cq(rx_depth + 1, NULL,
                                 ctx->channel, 0);

    {
        ibv::qp_init_attr attr;
        attr.send_cq = ctx->cq;
        attr.recv_cq = ctx->cq;
        attr.cap.max_send_wr  = 1;
        attr.cap.max_recv_wr  = rx_depth;
        attr.cap.max_send_sge = 1;
        attr.cap.max_recv_sge = 1;
        attr.qp_type = IBV_QPT_RC;

        ctx->qp = pd->create_qp(&attr);
    }

    {
        ibv::qp_attr attr;
        ibv::clear(attr);

        attr.qp_state        = IBV_QPS_INIT;
        attr.pkey_index      = 0;
        attr.port_num        = port->get_num();
        attr.qp_access_flags = 0;

        ctx->qp->modify(&attr,
                        ibv_qp_attr_mask(
                                IBV_QP_STATE              |
                                IBV_QP_PKEY_INDEX         |
                                IBV_QP_PORT               |
                                IBV_QP_ACCESS_FLAGS));
    }

    return ctx;
}

static int pp_post_recv(pingpong_context *ctx, int n)
{
    ibv::sge list;
    list.addr 	= (uintptr_t) ctx->buf;
    list.length = ctx->size;
    list.lkey 	= ctx->mr->get()->lkey;

    ibv::recv_wr wr;
    ibv::clear(wr);
    wr.wr_id      = PINGPONG_RECV_WRID;
    wr.sg_list    = &list;
    wr.num_sge    = 1;

    ibv::recv_wr *bad_wr;
    int i;

    for (i = 0; i < n; ++i)
        if (ctx->qp->post_recv(&wr, &bad_wr))
            break;

    return i;
}

static int pp_post_send(pingpong_context *ctx)
{
    ibv::sge list;
    list.addr 	= (uintptr_t) ctx->buf;
    list.length = ctx->size;
    list.lkey 	= ctx->mr->get()->lkey;

    ibv::send_wr wr;
    ibv::clear(wr);
    wr.wr_id 	    = PINGPONG_SEND_WRID;
    wr.sg_list    = &list;
    wr.num_sge    = 1;
    wr.opcode     = IBV_WR_SEND;
    wr.send_flags = IBV_SEND_SIGNALED;

    ibv::send_wr *bad_wr;

    return ctx->qp->post_send(&wr, &bad_wr);
}

void usage(const char *argv0)
{
	printf("Usage:\n");
	printf("  %s            start a server and wait for connection\n", argv0);
	printf("  %s <host>     connect to server at <host>\n", argv0);
	printf("\n");
	printf("Options:\n");
	printf("  -p, --port=<port>      listen on/connect to port <port> (default 18515)\n");
	printf("  -d, --ib-dev=<dev>     use IB device <dev> (default first device found)\n");
	printf("  -i, --ib-port=<port>   use port <port> of IB device (default 1)\n");
	printf("  -s, --size=<size>      size of message to exchange (default 4096)\n");
	printf("  -m, --mtu=<size>       path MTU (default 1024)\n");
	printf("  -r, --rx-depth=<dep>   number of receives to post at a time (default 500)\n");
	printf("  -n, --iters=<iters>    number of exchanges (default 1000)\n");
	printf("  -l, --sl=<sl>          service level value\n");
	printf("  -e, --events           sleep on CQ events (default poll)\n");
}

int main(int argc, char *argv[])
{
    pingpong_context *ctx;
    pingpong_dest     my_dest;
    shared_ptr<pingpong_dest> rem_dest;
	struct timeval           start, end;
    string                   ib_devname;
    string                   servername;
	int                      port = 18515;
	int                      ib_port = 1;
	int                      size = 4096;
	enum ibv_mtu		 mtu = IBV_MTU_1024;
	int                      rx_depth = 500;
	int                      iters = 1000;
	int                      use_event = 0;
	int                      routs;
	int                      rcnt, scnt;
	int                      num_cq_events = 0;
	int                      sl = 0;

	srand48(getpid() * time(NULL));

	while (1) {
		int c;

		static struct option long_options[] = {
			{ "port",     1, 0, 'p' },
			{ "ib-dev",   1, 0, 'd' },
			{ "ib-port",  1, 0, 'i' },
			{ "size",     1, 0, 's' },
			{ "mtu",      1, 0, 'm' },
			{ "rx-depth", 1, 0, 'r' },
			{ "iters",    1, 0, 'n' },
			{ "sl",       1, 0, 'l' },
			{ "events",   0, 0, 'e' },
			{ 0, 0, 0, 0 }
		};

		c = getopt_long(argc, argv, "p:d:i:s:m:r:n:l:e", long_options, NULL);
		if (c == -1)
			break;

		switch (c) {
		case 'p':
			port = strtol(optarg, NULL, 0);
			if (port < 0 || port > 65535) {
				usage(argv[0]);
				return 1;
			}
			break;

		case 'd':
			ib_devname = optarg;
			break;

		case 'i':
			ib_port = strtol(optarg, NULL, 0);
			if (ib_port < 0) {
				usage(argv[0]);
				return 1;
			}
			break;

		case 's':
			size = strtol(optarg, NULL, 0);
			break;

		case 'm':
			mtu = ibv::mtu_to_enum(strtol(optarg, NULL, 0));
			if (mtu < 0) {
				usage(argv[0]);
				return 1;
			}

		case 'r':
			rx_depth = strtol(optarg, NULL, 0);
			break;

		case 'n':
			iters = strtol(optarg, NULL, 0);
			break;

		case 'l':
			sl = strtol(optarg, NULL, 0);
			break;

		case 'e':
			++use_event;
			break;

		default:
			usage(argv[0]);
			return 1;
		}
	}

 	if (optind == argc - 1)
		servername = argv[optind];
	else if (optind < argc) {
		usage(argv[0]);
		return 1;
	}

	page_size = sysconf(_SC_PAGESIZE);

    try {
        ibv::port_ptr port;

        {
            ibv::device_list_ptr dev_list = ibv::get_device_list();
            ibv::device_list::iterator dev_it = dev_list->begin();

            if (!ib_devname.empty()) {
                for (dev_it=dev_list->begin(); dev_it!=dev_list->end(); ++dev_it)
                    if (dev_list->get_device_name(dev_it) == ib_devname)
                        break;
            }

            if (dev_it == dev_list->end()) {
                if (ib_devname.empty())
                    cerr << "No IB devices found" << endl;
                else
                    cerr << "IB device " << ib_devname << " not found" << endl;
                return 1;
            }

            ibv::context_ptr ib_ctx = dev_list->open_device(dev_it);
            port = ib_ctx->get_port(ib_port);
        }

        ctx = pp_init_ctx(port, size, rx_depth, use_event);

        routs = pp_post_recv(ctx, ctx->rx_depth);
        if (routs < ctx->rx_depth) {
            fprintf(stderr, "Couldn't post receive (%d)\n", routs);
            return 1;
        }

        if (use_event)
            ctx->cq->req_notify(false);

        my_dest.lid = port->get_lid();
        my_dest.qpn = ctx->qp->get_num();
        my_dest.psn = lrand48() & 0xffffff;
    } catch (ibv::error &e) {
        fprintf(stderr,"error: %s\n", e.what());
        if (e.errcode != 0)
            fprintf(stderr,"verbs error value: %d\n", e.errcode);
        return 1;
    } catch (std::exception &e) {
        fprintf(stderr,"error: %s\n", e.what());
        return 1;
    }

    printf("  local address:  LID 0x%04x, QPN 0x%06x, PSN 0x%06x\n",
           my_dest.lid, my_dest.qpn, my_dest.psn);

    try {
        if (!servername.empty())
            rem_dest = pp_client_exch_dest(servername.c_str(), port,
                                           &my_dest);
        else
            rem_dest = pp_server_exch_dest(ctx, ib_port, mtu, port,
                                           &my_dest);

        printf("  remote address: LID 0x%04x, QPN 0x%06x, PSN 0x%06x\n",
               rem_dest->lid, rem_dest->qpn, rem_dest->psn);

        if (!servername.empty())
            pp_connect_ctx(ctx, ib_port, my_dest.psn, mtu, rem_dest.get());
    } catch (ibv::error &e) {
        fprintf(stderr,"error: %s\n", e.what());
        if (e.errcode != 0)
            fprintf(stderr,"verbs error value: %d\n", e.errcode);
        return 1;
    } catch (std::exception &e) {
        fprintf(stderr,"error: %s\n", e.what());
        return 1;
    }

    ctx->pending = PINGPONG_RECV_WRID;

    if (!servername.empty()) {
        if (pp_post_send(ctx)) {
            fprintf(stderr, "Couldn't post send\n");
            return 1;
        }
        ctx->pending |= PINGPONG_SEND_WRID;
    }

    if (gettimeofday(&start, NULL)) {
        perror("gettimeofday");
        return 1;
    }

    rcnt = scnt = 0;
    while (rcnt < iters || scnt < iters) {
        try {
            if (use_event) {
                void *ev_ctx;

                ctx->channel->get_event(&ev_ctx);
                ctx->cq->req_notify(false);

                ++num_cq_events;
            }

            struct ibv_wc wc[2];
            int ne, i;

            do {
                ne = ctx->cq->poll(2, wc);
                if (ne < 0) {
                    fprintf(stderr, "poll CQ failed %d\n", ne);
                    return 1;
                }
            } while (!use_event && ne < 1);

            for (i = 0; i < ne; ++i) {
                if (wc[i].status != IBV_WC_SUCCESS) {
                    fprintf(stderr, "Failed status %d for wr_id %d\n",
                            wc[i].status, (int) wc[i].wr_id);
                    return 1;
                }

                switch ((int) wc[i].wr_id) {
                    case PINGPONG_SEND_WRID:
                        ++scnt;
                        break;

                    case PINGPONG_RECV_WRID:
                        if (--routs <= 1) {
                            routs += pp_post_recv(ctx, ctx->rx_depth - routs);
                            if (routs < ctx->rx_depth) {
                                fprintf(stderr,
                                        "Couldn't post receive (%d)\n",
                                        routs);
                                return 1;
                            }
                        }

                        ++rcnt;
                        break;

                    default:
                        fprintf(stderr, "Completion for unknown wr_id %d\n",
                                (int) wc[i].wr_id);
                        return 1;
                }

                ctx->pending &= ~(int) wc[i].wr_id;
                if (scnt < iters && !ctx->pending) {
                    if (pp_post_send(ctx)) {
                        fprintf(stderr, "Couldn't post send\n");
                        return 1;
                    }
                    ctx->pending = PINGPONG_RECV_WRID |
                        PINGPONG_SEND_WRID;
                }
            }
        } catch (ibv::error &e) {
            fprintf(stderr,"error: %s\n", e.what());
            if (e.errcode != 0)
                fprintf(stderr,"verbs error value: %d\n", e.errcode);
            return 1;
        } catch (std::exception &e) {
            fprintf(stderr,"error: %s\n", e.what());
            return 1;
        }
    }

    if (gettimeofday(&end, NULL)) {
        perror("gettimeofday");
        return 1;
    }

    {
        float usec = (end.tv_sec - start.tv_sec) * 1000000 +
            (end.tv_usec - start.tv_usec);
        long long bytes = (long long) size * iters * 2;

        printf("%lld bytes in %.2f seconds = %.2f Mbit/sec\n",
               bytes, usec / 1000000., bytes * 8. / usec);
        printf("%d iters in %.2f seconds = %.2f usec/iter\n",
               iters, usec / 1000000., usec / iters);
    }

    ctx->cq->ack_events(num_cq_events);

    return 0;
}
