/*
 * Copyright (c) 2008 Topspin Communications.  All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 * Fork Id: srq_pingpong.c 5257 2006-02-01 16:56:45Z roland
 */
/*
 * Copyright (C) 2008 United States Government as represented by the
 * Administrator of the National Aeronautics and Space Administration.
 * All Rights Reserved.
 *
 * $Id: srq_pingpong.cxx,v 1.8 2008/05/11 21:24:29 bgreen Exp $
 */

#if HAVE_CONFIG_H
#  include <config.h>
#endif /* HAVE_CONFIG_H */

#include <ibvpp.h>

#include <iostream>
#include <stdexcept>

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netdb.h>
#include <malloc.h>
#include <getopt.h>
#include <arpa/inet.h>
#include <time.h>

using namespace std;

enum {
	PINGPONG_RECV_WRID = 1,
	PINGPONG_SEND_WRID = 2,

	MAX_QP             = 256,
};

static int page_size;

struct pingpong_context {

	ibv::comp_channel_ptr   channel;

	ibv::mr_ptr             mr;
	ibv::cq_ptr             cq;
	ibv::srq_ptr            srq;
	ibv::qp_ptr             qp[MAX_QP];
	void			*buf;
	int			 size;
	unsigned		 num_qp;
	unsigned		 rx_depth;
	int			 pending[MAX_QP];
};

struct pingpong_dest {
	int lid;
	int qpn;
	int psn;
};

static void pp_connect_ctx(struct pingpong_context *ctx, int port, enum ibv_mtu mtu,
			  int sl, const struct pingpong_dest *my_dest,
			  const struct pingpong_dest *dest)
{
	int i;

	for (i = 0; i < ctx->num_qp; ++i) {
		ibv_qp_attr attr;
		ibv::clear(attr);
		attr.qp_state           = IBV_QPS_RTR;
		attr.path_mtu           = mtu;
		attr.dest_qp_num        = dest[i].qpn;
		attr.rq_psn             = dest[i].psn;
		attr.max_dest_rd_atomic	= 1;
		attr.min_rnr_timer	= 12;
		attr.ah_attr.is_global	= 0;
		attr.ah_attr.dlid	= dest[i].lid;
		attr.ah_attr.sl		= sl;
		attr.ah_attr.src_path_bits = 0;
		attr.ah_attr.port_num   = port;

		try {
		    ctx->qp[i]->modify(&attr,
		                       ibv_qp_attr_mask(
		                               IBV_QP_STATE              |
		                               IBV_QP_AV                 |
		                               IBV_QP_PATH_MTU           |
		                               IBV_QP_DEST_QPN           |
		                               IBV_QP_RQ_PSN             |
		                               IBV_QP_MAX_DEST_RD_ATOMIC |
		                               IBV_QP_MIN_RNR_TIMER));
		} catch (std::exception &e) {
		    fprintf(stderr, "Failed to modify QP[%d] to RTR\n", i);
		    throw;
		}

		attr.qp_state 	    = IBV_QPS_RTS;
		attr.timeout 	    = 14;
		attr.retry_cnt 	    = 7;
		attr.rnr_retry 	    = 7;
		attr.sq_psn 	    = my_dest[i].psn;
		attr.max_rd_atomic  = 1;
		try {
		    ctx->qp[i]->modify(&attr,
		                       ibv_qp_attr_mask(
		                               IBV_QP_STATE              |
		                               IBV_QP_TIMEOUT            |
		                               IBV_QP_RETRY_CNT          |
		                               IBV_QP_RNR_RETRY          |
		                               IBV_QP_SQ_PSN             |
		                               IBV_QP_MAX_QP_RD_ATOMIC));
		} catch (std::exception &e) {
		    fprintf(stderr, "Failed to modify QP[%d] to RTS\n", i);
		    throw;
		}
	}

	return;
}

static struct pingpong_dest *pp_client_exch_dest(const char *servername, int port,
						 const struct pingpong_dest *my_dest)
{
	struct addrinfo *res, *t;
	struct addrinfo hints;
	memset(&hints,0,sizeof(hints));
	hints.ai_family   = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	char *service;
	char msg[ sizeof "0000:000000:000000"];
	int n;
	int r;
	int i;
	int sockfd = -1;
	struct pingpong_dest *rem_dest = NULL;

	if (asprintf(&service, "%d", port) < 0)
		throw runtime_error("asprintf()");

	n = getaddrinfo(servername, service, &hints, &res);

	if (n < 0) {
		fprintf(stderr, "%s for %s:%d\n", gai_strerror(n), servername, port);
		free(service);
		throw runtime_error("getaddrinfo()");
	}

	for (t = res; t; t = t->ai_next) {
		sockfd = socket(t->ai_family, t->ai_socktype, t->ai_protocol);
		if (sockfd >= 0) {
			if (!connect(sockfd, t->ai_addr, t->ai_addrlen))
				break;
			close(sockfd);
			sockfd = -1;
		}
	}

	freeaddrinfo(res);
	free(service);

	if (sockfd < 0) {
		fprintf(stderr, "Couldn't connect to %s:%d\n", servername, port);
		throw runtime_error("connect()");
	}

	for (i = 0; i < MAX_QP; ++i) {
		sprintf(msg, "%04x:%06x:%06x", my_dest[i].lid, my_dest[i].qpn, my_dest[i].psn);
		if (write(sockfd, msg, sizeof msg) != sizeof msg) {
			fprintf(stderr, "Couldn't send local address\n");
			close(sockfd);
			throw runtime_error("write()");
		}
	}

	try {
		rem_dest = new pingpong_dest[MAX_QP];
	} catch (...) {
		close(sockfd);
		throw;
	}

    try {

	for (i = 0; i < MAX_QP; ++i) {
		n = 0;
		while (n < sizeof msg) {
			r = read(sockfd, msg + n, sizeof msg - n);
			if (r < 0) {
				perror("client read");
				fprintf(stderr, "%d/%d: Couldn't read remote address [%d]\n",
					n, (int) sizeof msg, i);
				throw runtime_error("read()");
			}
			n += r;
		}

		sscanf(msg, "%x:%x:%x",
		       &rem_dest[i].lid, &rem_dest[i].qpn, &rem_dest[i].psn);
	}
    } catch (...) {
        close(sockfd);
        if (rem_dest) delete[] rem_dest;
        throw;
    }

	write(sockfd, "done", sizeof "done");
	close(sockfd);

	return rem_dest;
}

static struct pingpong_dest *pp_server_exch_dest(struct pingpong_context *ctx,
						 int ib_port, enum ibv_mtu mtu,
						 int port, int sl,
						 const struct pingpong_dest *my_dest)
{
	struct addrinfo *res, *t;
	struct addrinfo hints;
	memset(&hints,0,sizeof(hints));
	hints.ai_flags    = AI_PASSIVE;
	hints.ai_family   = AF_UNSPEC;
	hints.ai_socktype = SOCK_STREAM;
	char *service;
	char msg[ sizeof "0000:000000:000000"];
	int n;
	int r;
	int i;
	int sockfd = -1, connfd;
	struct pingpong_dest *rem_dest = NULL;

	if (asprintf(&service, "%d", port) < 0)
		throw runtime_error("asprintf()");

	n = getaddrinfo(NULL, service, &hints, &res);

	if (n < 0) {
		fprintf(stderr, "%s for port %d\n", gai_strerror(n), port);
		free(service);
		throw runtime_error("getaddrinfo()");
	}

	for (t = res; t; t = t->ai_next) {
		sockfd = socket(t->ai_family, t->ai_socktype, t->ai_protocol);
		if (sockfd >= 0) {
			n = 1;

			setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &n, sizeof n);

			if (!bind(sockfd, t->ai_addr, t->ai_addrlen))
				break;
			close(sockfd);
			sockfd = -1;
		}
	}

	freeaddrinfo(res);
	free(service);

	if (sockfd < 0) {
		fprintf(stderr, "Couldn't listen to port %d\n", port);
		throw runtime_error("bind()");
	}

    try {
	listen(sockfd, 1);
	connfd = accept(sockfd, NULL, 0);
	close(sockfd);
	if (connfd < 0) {
		fprintf(stderr, "accept() failed\n");
		throw runtime_error("accept()");
	}

	rem_dest = new pingpong_dest[MAX_QP];

	for (i = 0; i < MAX_QP; ++i) {
		n = 0;
		while (n < sizeof msg) {
			r = read(connfd, msg + n, sizeof msg - n);
			if (r < 0) {
				perror("server read");
				fprintf(stderr, "%d/%d: Couldn't read remote address [%d]\n",
					n, (int) sizeof msg, i);
				throw runtime_error("read()\n");
			}
			n += r;
		}

		sscanf(msg, "%x:%x:%x",
		       &rem_dest[i].lid, &rem_dest[i].qpn, &rem_dest[i].psn);
	}

	try {
	    pp_connect_ctx(ctx, ib_port, mtu, sl, my_dest, rem_dest);
	} catch (std::exception &e) {
	    fprintf(stderr, "Couldn't connect to remote QP: %s\n", e.what());
	    throw;
	}

	for (i = 0; i < MAX_QP; ++i) {
		sprintf(msg, "%04x:%06x:%06x", my_dest[i].lid, my_dest[i].qpn, my_dest[i].psn);
		if (write(connfd, msg, sizeof msg) != sizeof msg) {
			fprintf(stderr, "Couldn't send local address\n");
			throw runtime_error("write()");
		}
	}

	read(connfd, msg, sizeof msg);
    } catch (...) {
        if (rem_dest) delete[] rem_dest;
        close(connfd);
        throw;
    }

	close(connfd);
	return rem_dest;
}

static struct pingpong_context *pp_init_ctx(const ibv::port_ptr &port, int size,
					    int num_qp, int rx_depth,
					    int use_event)
{
	pingpong_context *ctx;
	int i;

	ctx = new pingpong_context();

	ctx->size     = size;
	ctx->num_qp   = num_qp;
	ctx->rx_depth = rx_depth;

	ctx->buf = memalign(page_size, size);
	if (!ctx->buf)
		throw runtime_error("Couldn't allocate work buf.");

	memset(ctx->buf, 0, size);

	ibv::context_ptr context = port->get_context();

	if (use_event)
		ctx->channel = context->create_comp_channel();

	ibv::pd_ptr pd = context->alloc_pd();

	ctx->mr = pd->reg_mr(ctx->buf, size, IBV_ACCESS_LOCAL_WRITE);

	ctx->cq = context->create_cq(rx_depth + num_qp, NULL,
	                             ctx->channel, 0);

	{
		ibv::srq_init_attr attr;
		ibv::clear(attr);
		attr.attr.max_wr  = rx_depth;
		attr.attr.max_sge = 1;

		ctx->srq = pd->create_srq(&attr);
	}

	for (i = 0; i < num_qp; ++i) {
		ibv::qp_init_attr attr;
		attr.send_cq = ctx->cq;
		attr.recv_cq = ctx->cq;
		attr.srq = ctx->srq;
		attr.cap.max_send_wr  = 1;
		attr.cap.max_send_sge = 1;
		attr.qp_type = IBV_QPT_RC;

		ctx->qp[i] = pd->create_qp(&attr);
	}

	for (i = 0; i < num_qp; ++i) {
		ibv::qp_attr attr;
		ibv::clear(attr);
		attr.qp_state        = IBV_QPS_INIT;
		attr.pkey_index      = 0;
		attr.port_num        = port->get_num();
		attr.qp_access_flags = 0;

		ctx->qp[i]->modify(&attr,
		                   ibv::qp_attr_mask(
		                           IBV_QP_STATE              |
		                           IBV_QP_PKEY_INDEX         |
		                           IBV_QP_PORT               |
		                           IBV_QP_ACCESS_FLAGS));
	}

	return ctx;
}

/* unneeded, because destructors do cleanup:
int pp_close_ctx(struct pingpong_context *ctx, int num_qp)
{
}
*/

static int pp_post_recv(struct pingpong_context *ctx, int n)
{
	ibv::sge list;
	list.addr 	= (uintptr_t) ctx->buf;
	list.length = ctx->size;
	list.lkey 	= ctx->mr->get()->lkey;

	ibv::recv_wr wr; ibv::clear(wr);
	wr.wr_id      = PINGPONG_RECV_WRID;
	wr.sg_list    = &list;
	wr.num_sge    = 1;

	ibv::recv_wr *bad_wr;
	int i;

	for (i = 0; i < n; ++i)
		if (ctx->srq->post_recv(&wr, &bad_wr))
			break;

	return i;
}

static int pp_post_send(struct pingpong_context *ctx, int qp_index)
{
	ibv::sge list;
	list.addr 	= (uintptr_t) ctx->buf;
	list.length = ctx->size;
	list.lkey 	= ctx->mr->get()->lkey;

	ibv::send_wr wr; ibv::clear(wr);
	wr.wr_id 	    = PINGPONG_SEND_WRID;
	wr.sg_list    = &list;
	wr.num_sge    = 1;
	wr.opcode     = IBV_WR_SEND;
	wr.send_flags = IBV_SEND_SIGNALED;

	ibv::send_wr *bad_wr;

	return ctx->qp[qp_index]->post_send(&wr, &bad_wr);
}

static int find_qp(int qpn, struct pingpong_context *ctx, int num_qp)
{
	int i;

	for (i = 0; i < num_qp; ++i)
		if (ctx->qp[i]->get_num() == qpn)
			return i;

	return -1;
}

static void usage(const char *argv0)
{
	printf("Usage:\n");
	printf("  %s            start a server and wait for connection\n", argv0);
	printf("  %s <host>     connect to server at <host>\n", argv0);
	printf("\n");
	printf("Options:\n");
	printf("  -p, --port=<port>      listen on/connect to port <port> (default 18515)\n");
	printf("  -d, --ib-dev=<dev>     use IB device <dev> (default first device found)\n");
	printf("  -i, --ib-port=<port>   use port <port> of IB device (default 1)\n");
	printf("  -s, --size=<size>      size of message to exchange (default 4096)\n");
	printf("  -m, --mtu=<size>       path MTU (default 1024)\n");
	printf("  -q, --num-qp=<num>     number of QPs to use (default 16)\n");
	printf("  -r, --rx-depth=<dep>   number of receives to post at a time (default 500)\n");
	printf("  -n, --iters=<iters>    number of exchanges per QP(default 1000)\n");
	printf("  -l, --sl=<sl>          service level value\n");
	printf("  -e, --events           sleep on CQ events (default poll)\n");
}

int main(int argc, char *argv[])
{
	ibv_wc                  *wc;
	struct pingpong_context *ctx;
	struct pingpong_dest     my_dest[MAX_QP];
	pingpong_dest           *rem_dest = 0;
	struct timeval           start, end;
	string                   ib_devname;
	string                   servername;
	int                      port = 18515;
	int                      ib_port = 1;
	int                      size = 4096;
	enum ibv_mtu		 mtu = IBV_MTU_1024;
	int                      num_qp = 16;
	int                      rx_depth = 500;
	int                      iters = 1000;
	int                      use_event = 0;
	int                      routs;
	int                      rcnt, scnt;
	int			 num_wc;
	int                      i;
	int                      num_cq_events = 0;
	int                      sl = 0;

	srand48(getpid() * time(NULL));

	while (1) {
		int c;

		static struct option long_options[] = {
			{ "port",     1, 0, 'p' },
			{ "ib-dev",   1, 0, 'd' },
			{ "ib-port",  1, 0, 'i' },
			{ "size",     1, 0, 's' },
			{ "mtu",      1, 0, 'm' },
			{ "num-qp",   1, 0, 'q' },
			{ "rx-depth", 1, 0, 'r' },
			{ "iters",    1, 0, 'n' },
			{ "sl",       1, 0, 'l' },
			{ "events",   0, 0, 'e' },
			{ 0 }
		};

		c = getopt_long(argc, argv, "p:d:i:s:m:q:r:n:l:e", long_options, NULL);
		if (c == -1)
			break;

		switch (c) {
		case 'p':
			port = strtol(optarg, NULL, 0);
			if (port < 0 || port > 65535) {
				usage(argv[0]);
				return 1;
			}
			break;

		case 'd':
			ib_devname = optarg;
			break;

		case 'i':
			ib_port = strtol(optarg, NULL, 0);
			if (ib_port < 0) {
				usage(argv[0]);
				return 1;
			}
			break;

		case 's':
			size = strtol(optarg, NULL, 0);
			break;

		case 'm':
			mtu = ibv::mtu_to_enum(strtol(optarg, NULL, 0));
			if (mtu < 0) {
				usage(argv[0]);
				return 1;
			}
			break;

		case 'q':
			num_qp = strtol(optarg, NULL, 0);
			break;

		case 'r':
			rx_depth = strtol(optarg, NULL, 0);
			break;

		case 'n':
			iters = strtol(optarg, NULL, 0);
			break;

		case 'l':
			sl = strtol(optarg, NULL, 0);
			break;

		case 'e':
			++use_event;
			break;

		default:
			usage(argv[0]);
			return 1;
		}
	}

	if (optind == argc - 1)
		servername = argv[optind];
	else if (optind < argc) {
		usage(argv[0]);
		return 1;
	}

	if (num_qp > rx_depth) {
		fprintf(stderr, "rx_depth %d is too small for %d QPs -- "
			"must have at least one receive per QP.\n",
			rx_depth, num_qp);
		return 1;
	}

	num_wc = num_qp + rx_depth;
	wc     = new ibv_wc[num_wc];

	page_size = sysconf(_SC_PAGESIZE);

    try {
	ibv::port_ptr port;

	{
	    ibv::device_list_ptr dev_list = ibv::get_device_list();
	    ibv::device_list::iterator dev_it = dev_list->begin();

	    if (!ib_devname.empty()) {
	        for (dev_it=dev_list->begin(); dev_it!=dev_list->end(); ++dev_it)
	            if (dev_list->get_device_name(dev_it) == ib_devname)
	                break;
	    }

	    if (dev_it == dev_list->end()) {
	        if (ib_devname.empty())
	            cerr << "No IB devices found" << endl;
	        else
	            cerr << "IB device " << ib_devname << " not found" << endl;
	        return 1;
	    }

	    ibv::context_ptr ib_ctx = dev_list->open_device(dev_it);
	    port = ib_ctx->get_port(ib_port);
	}

	ctx = pp_init_ctx(port, size, num_qp, rx_depth, use_event);

	routs = pp_post_recv(ctx, ctx->rx_depth);
	if (routs < ctx->rx_depth) {
		fprintf(stderr, "Couldn't post receive (%d)\n", routs);
		return 1;
	}

	if (use_event)
		ctx->cq->req_notify(false);

	for (i = 0; i < num_qp; ++i) {
		my_dest[i].qpn = ctx->qp[i]->get_num();
		my_dest[i].psn = lrand48() & 0xffffff;
		my_dest[i].lid = port->get_lid();

		printf("  local address:  LID 0x%04x, QPN 0x%06x, PSN 0x%06x\n",
		       my_dest[i].lid, my_dest[i].qpn, my_dest[i].psn);
	}
    } catch (ibv::error &e) {
        fprintf(stderr,"error: %s\n", e.what());
        if (e.errcode != 0)
            fprintf(stderr,"verbs error value: %d\n", e.errcode);
        return 1;
    } catch (std::exception &e) {
        fprintf(stderr,"error: %s\n", e.what());
        return 1;
    }

    try {
	if (!servername.empty())
		rem_dest = pp_client_exch_dest(servername.c_str(), port, my_dest);
	else
		rem_dest = pp_server_exch_dest(ctx, ib_port, mtu, port, sl, my_dest);

	for (i = 0; i < num_qp; ++i)
		printf("  remote address: LID 0x%04x, QPN 0x%06x, PSN 0x%06x\n",
		       rem_dest[i].lid, rem_dest[i].qpn, rem_dest[i].psn);

	if (!servername.empty())
		pp_connect_ctx(ctx, ib_port, mtu, sl, my_dest, rem_dest);
    } catch (ibv::error &e) {
        fprintf(stderr,"error: %s\n", e.what());
        if (e.errcode != 0)
            fprintf(stderr,"verbs error value: %d\n", e.errcode);
        if (rem_dest) delete[] rem_dest;
        return 1;
    } catch (std::exception &e) {
        fprintf(stderr,"error: %s\n", e.what());
        return 1;
    }

	if (!servername.empty())
		for (i = 0; i < num_qp; ++i) {
			if (pp_post_send(ctx, i)) {
				fprintf(stderr, "Couldn't post send\n");
				return 1;
			}
			ctx->pending[i] = PINGPONG_SEND_WRID | PINGPONG_RECV_WRID;
		}
	else
		for (i = 0; i < num_qp; ++i)
			ctx->pending[i] = PINGPONG_RECV_WRID;

	if (gettimeofday(&start, NULL)) {
		perror("gettimeofday");
		return 1;
	}

	rcnt = scnt = 0;
	while (rcnt < iters || scnt < iters) {
	    try {
		if (use_event) {
			void *ev_ctx;

			ctx->channel->get_event(&ev_ctx);

			++num_cq_events;

			ctx->cq->req_notify(false);
		}

		{
			int ne, qp_ind;

			do {
				ne = ctx->cq->poll(num_wc, wc);
				if (ne < 0) {
					fprintf(stderr, "poll CQ failed %d\n", ne);
					return 1;
				}
			} while (!use_event && ne < 1);

			for (i = 0; i < ne; ++i) {
				if (wc[i].status != IBV_WC_SUCCESS) {
#if HAVE_IBV_WC_STATUS_STR
					fprintf(stderr, "Failed status %s (%d) for wr_id %d\n",
						ibv_wc_status_str(wc[i].status),
						wc[i].status, (int) wc[i].wr_id);
#else
					fprintf(stderr, "Failed status %d for wr_id %d\n",
						wc[i].status, (int) wc[i].wr_id);
#endif
					return 1;
				}

				qp_ind = find_qp(wc[i].qp_num, ctx, num_qp);
				if (qp_ind < 0) {
					fprintf(stderr, "Couldn't find QPN %06x\n",
						wc[i].qp_num);
					return 1;
				}

				switch ((int) wc[i].wr_id) {
				case PINGPONG_SEND_WRID:
					++scnt;
					break;

				case PINGPONG_RECV_WRID:
					if (--routs <= num_qp) {
						routs += pp_post_recv(ctx, ctx->rx_depth - routs);
						if (routs < ctx->rx_depth) {
							fprintf(stderr,
								"Couldn't post receive (%d)\n",
								routs);
							return 1;
						}
					}

					++rcnt;
					break;

				default:
					fprintf(stderr, "Completion for unknown wr_id %d\n",
						(int) wc[i].wr_id);
					return 1;
				}

				ctx->pending[qp_ind] &= ~(int) wc[i].wr_id;
				if (scnt < iters && !ctx->pending[qp_ind]) {
					if (pp_post_send(ctx, qp_ind)) {
						fprintf(stderr, "Couldn't post send\n");
						return 1;
					}
					ctx->pending[qp_ind] = PINGPONG_RECV_WRID |
							       PINGPONG_SEND_WRID;
				}

			}
		}
	    } catch (ibv::error &e) {
                fprintf(stderr,"error: %s\n", e.what());
                if (e.errcode != 0)
                    fprintf(stderr,"verbs error value: %d\n", e.errcode);
                return 1;
            } catch (std::exception &e) {
                fprintf(stderr,"error: %s\n", e.what());
                return 1;
            }

	}

	if (gettimeofday(&end, NULL)) {
		perror("gettimeofday");
		return 1;
	}

	{
		float usec = (end.tv_sec - start.tv_sec) * 1000000 +
			(end.tv_usec - start.tv_usec);
		long long bytes = (long long) size * iters * 2;

		printf("%lld bytes in %.2f seconds = %.2f Mbit/sec\n",
		       bytes, usec / 1000000., bytes * 8. / usec);
		printf("%d iters in %.2f seconds = %.2f usec/iter\n",
		       iters, usec / 1000000., usec / iters);
	}

	ctx->cq->ack_events(num_cq_events);

	return 0;
}
