/*
 * Copyright (C) 2008 United States Government as represented by the
 * Administrator of the National Aeronautics and Space Administration.
 * All Rights Reserved.
 */

#include "ibvpp.h"

#ifdef HAVE_PTHREAD_H
#include <pthread.h>
#endif
#include <stdexcept>
#include <sstream>
#include <assert.h>
#include <errno.h>

using IBVPP_SPNS::const_pointer_cast;

#if DEBUG
#define DEBUG_MESG(f,v) printf(f " \n",v)
#else
#define DEBUG_MESG(f,v)
#endif

extern "C" {
    void ibvpp_is_here() {
        // does nothing but present an unmangled symbol for lookup.
    }
}

namespace ibv {

    namespace {
#ifdef HAVE_PTHREAD_MUTEX_LOCK
        pthread_mutex_t ibv_mutex = PTHREAD_MUTEX_INITIALIZER;
#define LOCK_IBV() pthread_mutex_lock(&ibv_mutex)
#define UNLOCK_IBV() pthread_mutex_unlock(&ibv_mutex)
#else
#define LOCK_IBV()
#define UNLOCK_IBV()
#endif
        struct scoped_ibv_lock {
            scoped_ibv_lock() { LOCK_IBV(); }
            ~scoped_ibv_lock() { UNLOCK_IBV(); }
        };
    }

    //-----------------------------------------------------------
    // ibv_context

    context::~context()
    {
        DEBUG_MESG("ibv_close_device(0x%x)",context_);
        ibv_close_device(context_);
        // how should failure be handled?
    }

    void context::query_device(device_attr *device_attr) const
        throw (error)
    {
        clear(*device_attr);
        int result = ibv_query_device(context_,device_attr);
        if (result != 0)
            throw error("ibv_query_device", result);
    }

    const_device_attr_ptr context::get_device_attrs() const
        throw (std::exception)
    {
        scoped_ibv_lock l;
        if (device_attrs == 0) {
            device_attrs = device_attr_ptr(new ibv_device_attr);
            query_device(device_attrs.get());
        }
        return device_attrs;
    }

    void context::query_port(uint8_t port_num, port_attr *port_attr) const
        throw (error)
    {
        clear(*port_attr);
        int result = ibv_query_port(context_,port_num,port_attr);
        if (result != 0)
            throw error("ibv_query_port", result);
    }

    void context::query_gid(uint8_t port_num, int index, ibv_gid *gid) const
        throw (error)
    {
        memset(gid,0,sizeof(ibv_gid));
        int result = ibv_query_gid(context_,port_num,index,gid);
        if (result != 0)
            throw error("ibv_query_gid", result);
    }

    void context::query_pkey(uint8_t port_num, int index, uint16_t *pkey) const
        throw (error)
    {
        int result = ibv_query_pkey(context_,port_num,index,pkey);
        if (result != 0)
            throw error("ibv_query_pkey", result);
    }

    port_ptr context::get_port(uint8_t port_num)
    {
        return port_ptr(new port(shared_from_this(), port_num));
    }

    const_port_ptr context::get_port(uint8_t port_num) const
    {
        return const_port_ptr(new port(const_pointer_cast<context>(shared_from_this()), port_num));
    }

    //-----------------------------------------------------------
    // port

    std::string port::get_name() const
    {
        std::ostringstream strm;
        strm << context_->get_device_name() << '.' << portnum_;
        return strm.str();
    }

    uint16_t port::get_lid() const
    {
        port_attr pattr;
        query(&pattr);
        return pattr.lid;
    }

    //-----------------------------------------------------------
    // ibv_pd

    pd::pd(const context_ptr &ctx) throw (std::exception)
        : context_(ctx)
    {
        pd_ = ibv_alloc_pd(context_->get());
        if (pd_ == 0)
            throw error("ibv_alloc_pd", errno);
        DEBUG_MESG("ibv_alloc_pd(0x%x)",pd_);
    }

    pd::~pd()
    {
        DEBUG_MESG("ibv_dealloc_pd(0x%x)",pd_);
        ibv_dealloc_pd(pd_);
        // how should failure be handled?
    }

    mr_ptr pd::reg_mr(void *addr,size_t length,
                      ibv_access_flags access)
    {
        mr_ptr mrp(new mr(shared_from_this(),addr,length,access));
        return mrp;
    }

    //-----------------------------------------------------------
    // ibv_mr

    mr::mr(const pd_ptr &pdp,void *addr,size_t length,
           enum ibv_access_flags access) throw (std::exception)
        : pd_(pdp)
    {
        mr_ = ibv_reg_mr(pd_->get(),addr,length,access);
        if (mr_ == 0)
            throw error("ibv_reg_mr", errno);
        DEBUG_MESG("ibv_reg_mr(0x%x)",mr_);
    }

    mr::~mr()
    {
        DEBUG_MESG("ibv_dereg_mr(0x%x)",mr_);
        ibv_dereg_mr(mr_);
        // how should failure be handled?
    }

    //-----------------------------------------------------------
    // ibv_srq

    srq::srq(const pd_ptr &pdp,srq_init_attr *init_attr)
        throw (error)
        : pd_(pdp)
    {
        srq_ = ibv_create_srq(pd_->get(), init_attr);
        if (srq_ == 0)
            throw error("ibv_create_srq", errno);
        DEBUG_MESG("ibv_create_srq(0x%x)",srq_);
    }

    srq::~srq()
    {
        DEBUG_MESG("ibv_destroy_srq(0x%x)",srq_);
        ibv_destroy_srq(srq_);
        // how should failure be handled?
    }

    void srq::modify(srq_attr *attr,srq_attr_mask attr_mask) throw (error)
    {
        int result = ibv_modify_srq(srq_,attr,attr_mask);
        if (result != 0)
            throw error("ibv_modify_srq", result);
    }

    void srq::query(srq_attr *attr) const throw (error)
    {
        int result = ibv_query_srq(srq_,attr);
        if (result != 0)
            throw error("ibv_query_srq", result);
    }

    //-----------------------------------------------------------
    // ibv_ah

    ah::ah(const pd_ptr &pdp,ah_attr *attr) throw (error)
        : pd_(pdp)
    {
        ah_ = ibv_create_ah(pd_->get(), attr);
        if (ah_ == 0)
            throw error("ibv_create_ah", errno);
        DEBUG_MESG("ibv_create_ah(0x%x)",ah_);
    }

    ah::~ah()
    {
        DEBUG_MESG("ibv_destroy_ah(0x%x)",ah_);
        ibv_destroy_ah(ah_);
        // how should failure be handled?
    }

    //-----------------------------------------------------------
    // ibv_qp

    qp::qp(const pd_ptr &pdp,qp_init_attr *init_attr)
        throw (error)
        : pd_(pdp)
    {
        ibv_qp_init_attr in_attr;
        in_attr.qp_context = init_attr->qp_context;
        if (init_attr->send_cq)
            in_attr.send_cq    = init_attr->send_cq->get();
        else
            in_attr.send_cq    = 0;
        if (init_attr->recv_cq)
            in_attr.recv_cq    = init_attr->recv_cq->get();
        else
            in_attr.recv_cq    = 0;
        if (init_attr->srq)
            in_attr.srq        = init_attr->srq->get();
        else
            in_attr.srq        = 0;
        in_attr.cap        = init_attr->cap;
        in_attr.qp_type    = init_attr->qp_type;
        in_attr.sq_sig_all = init_attr->sq_sig_all;

        scq_  = init_attr->send_cq;
        rcq_  = init_attr->recv_cq;
        srq_  = init_attr->srq;

        qp_ = ibv_create_qp(pd_->get(), &in_attr);
        if (qp_ == 0)
            throw error("ibv_create_qp", errno);

        DEBUG_MESG("ibv_create_qp(0x%x)",qp_);
    }

    qp::~qp()
    {
        DEBUG_MESG("ibv_destroy_qp(0x%x)",qp_);
        ibv_destroy_qp(qp_);
        // how should failure be handled?
    }

    void qp::modify(qp_attr *attr,qp_attr_mask attr_mask) throw (error)
    {
        int result = ibv_modify_qp(qp_,attr,attr_mask);
        if (result != 0)
            throw error("ibv_modify_qp", result);
    }

    void qp::query(qp_attr *attr,
                   qp_attr_mask attr_mask,
                   qp_init_attr *init_attr) const throw (error)
    {
        clear(*attr);
        ibv_qp_init_attr out_init_attr;
        int result = ibv_query_qp(qp_,attr,attr_mask,&out_init_attr);
        if (result != 0)
            throw error("ibv_query_qp", result);

        init_attr->qp_context = out_init_attr.qp_context;
        init_attr->send_cq    = scq_;
        init_attr->recv_cq    = rcq_;
        init_attr->srq        = srq_;
        init_attr->cap        = out_init_attr.cap;
        init_attr->qp_type    = out_init_attr.qp_type;
        init_attr->sq_sig_all = out_init_attr.sq_sig_all;
    }

    //-----------------------------------------------------------
    // ibv_comp_channel

    comp_channel::comp_channel(const context_ptr &ctx)
        : context_(ctx)
    {
        comp_channel_ = ibv_create_comp_channel(context_->get());
        if (comp_channel_ == 0)
            throw error("ibv_create_comp_channel", errno);
        DEBUG_MESG("ibv_create_comp_channel(0x%x)",comp_channel_);
    }

    comp_channel::~comp_channel()
    {
        DEBUG_MESG("ibv_destroy_comp_channel(0x%x)",comp_channel_);
        ibv_destroy_comp_channel(comp_channel_);
        // how should failure be handled?
    }

    cq_ptr comp_channel::get_event(void **cq_context) throw (error)
    {
        ibv_cq *event_cq;
        int result = ibv_get_cq_event(comp_channel_,&event_cq,cq_context);
        if (result != 0)
            throw error("ibv_get_cq_event", result);

        scoped_ibv_lock l;

        cq_map_type::iterator it = cq_map.find(event_cq->handle);
        if (it == cq_map.end())
            throw error("ibv_get_cq_event(): received a cq event for an unknown cq");

        cq_ptr cq((*it).second);
        if (cq == 0) {
            cq_map.erase(it);
            throw error("ibv_get_cq_event(): received a cq event for an expired cq");
        }

        return cq;
    }

    //-----------------------------------------------------------
    // ibv_cq

    cq_ptr context::create_cq(int cqe,void *cq_context,
                              const comp_channel_ptr &cchan,
                              int comp_vector) throw (std::exception)
    {
        ibv_comp_channel *ccp = 0;
        if (cchan) ccp = cchan->get();

        context_ptr ctx(shared_from_this());

        ibv_cq *cq_ = ibv_create_cq(ctx->get(),cqe,cq_context,
                                    ccp, comp_vector);

        if (cq_ == 0)
            throw error("ibv_create_cq", errno);

        DEBUG_MESG("ibv_create_cq(0x%x)",cq_);

        cq_ptr cqp(new cq(ctx,cq_));

        if (cchan) {
            scoped_ibv_lock l;
            cchan->cq_map[cq_->handle] = cqp;
            cqp->comp_channel_ = cchan;
        }

        return cqp;
    }

    cq::~cq()
    {
        comp_channel_ptr cchan = comp_channel_.lock();
        if (cchan) {
            scoped_ibv_lock l;
            comp_channel::cq_map_type::iterator it =
                cchan->cq_map.find(cq_->handle);
            if (it != cchan->cq_map.end())
                cchan->cq_map.erase(it);
        }

        DEBUG_MESG("ibv_destroy_cq(0x%x)",cq_);
        ibv_destroy_cq(cq_);
        // how should failure be handled?
    }

    void cq::resize(int cqe) throw (error)
    {
        int result = ibv_resize_cq(cq_,cqe);
        if (result != 0)
            throw error("ibv_resize_cq", result);
    }

    void cq::ack_events(unsigned int nevents) throw ()
    {
        ibv_ack_cq_events(cq_,nevents);
    }

    unsigned cq::poll(int num_entries,wc *wc) throw (error) {
        int result = ibv_poll_cq(cq_,num_entries,wc);
        if (result == -1)
            throw error("ibv_poll_cq", result);
        assert(result >= 0);
        return unsigned(result);
    }

    void cq::req_notify(bool solicited_only) throw (error)
    {
        int result = ibv_req_notify_cq(cq_,solicited_only);
        if (result != 0)
            throw error("ibv_req_notify_cq", result);
    }

    //-----------------------------------------------------------
    // ibv_device

    namespace {
        inline context_ptr open_device(ibv_device *dev)
            throw (std::exception)
        {
            ibv_context *ctx = ibv_open_device(dev);
            if (ctx == 0)
                throw error("ibv_open_device", errno);
            DEBUG_MESG("ibv_open_device(0x%x)",ctx);
            return context_ptr(new context(ctx));
        }
    }


    context_ptr device::open_device() throw (std::exception)
    {
        return ibv::open_device(device_);
    }

    //-----------------------------------------------------------
    // ibv_device_list

    namespace {
        IBVPP_SPNS::weak_ptr<device_list> singleton_device_list;
    }

    device_list::~device_list() throw ()
    {
        DEBUG_MESG("ibv_free_device_list(0x%x)",device_list_);
        ibv_free_device_list(device_list_);
    }

    device_list_ptr get_device_list() throw (std::exception)
    {
        device_list_ptr result;

        // Ensure that only one device list is ever allocated at a time.
        // Invariant: 'singleton_device_list.lock()' returns the allocated
        // device_list, or '0' if a device_list is not allocated.

        scoped_ibv_lock l;

        result = singleton_device_list.lock();
        if (result == 0) {
            ibvpp_is_here();
            int num_devices;
            ibv_device **devlist = ibv_get_device_list(&num_devices);
            if (devlist == 0) {
                UNLOCK_IBV();
                throw error("ibv_get_device_list", errno);
            }
            DEBUG_MESG("ibv_get_device_list(0x%x)",devlist);
            result = device_list_ptr(new device_list(devlist,num_devices));
            singleton_device_list = IBVPP_SPNS::weak_ptr<device_list>(result);
        }

        return result;
    }
 
    context_ptr device_list::open_device(const iterator &it)
        throw (std::exception)
    {
        return ibv::open_device(*it);
    }

    //-----------------------------------------------------------
    // port iterator

    class port_iterator::detail
    {
    public:
        static void check_device(Impl_ &impl_);
        static void increment(Impl_ &impl_);
        static void decrement(Impl_ &impl_);
    };

    void port_iterator::detail::check_device(Impl_ &impl_)
    {
        if (impl_.context_ == 0) {
            ibv_context *ctx = ibv_open_device(*impl_.pos_.first);
            if (ctx == 0)
                throw error("ibv_open_device", errno);
            impl_.context_ = context_ptr(new context(ctx));
        }
    }

    void port_iterator::detail::increment(Impl_ &impl_)
    {
        // increment the port number;  if its the last port of the device,
        // jump to next device.
        detail::check_device(impl_);
        if (++impl_.pos_.second ==
            impl_.context_->get_device_attrs()->phys_port_cnt)
        {
            ++impl_.pos_.first;      // next device
            impl_.pos_.second = 0;   // first port of new device
            impl_.context_.reset();  // invalidate context;
                                     // new device will be queried on demand.
            impl_.port_.reset();
        }
    }

    void port_iterator::detail::decrement(Impl_ &impl_)
    {
        // if its port 0, jump to the last port of the previous device;
        // otherwise, decrement the port number.
        if (impl_.pos_.second == 0) {
            --impl_.pos_.first;      // previous device
            impl_.context_.reset();  // invalidate context
            impl_.port_.reset();
            // we need to query the device to get its port count.
            check_device(impl_);
            impl_.pos_.second =
                impl_.context_->get_device_attrs()->phys_port_cnt - 1;
        }
        else
            --impl_.pos_.second;
    }

    port_iterator::port_iterator(const device_list::iterator &dit,uint8_t pnum)
        : impl_(dit,pnum)
    {
    }

    port_ptr port_iterator::operator*() const {
        if (impl_.port_ == 0) {
            detail::check_device(impl_);
            impl_.port_ = port_ptr(new port(impl_.context_,
                                            impl_.pos_.second+1));
        }

        return impl_.port_;
    }

    port_iterator &port_iterator::operator++() {
        detail::increment(impl_);
        return *this;
    }

    port_iterator port_iterator::operator++(int) {
        port_iterator it(*this);
        detail::increment(impl_);
        return it;
    }

    port_iterator &port_iterator::operator--() {
        detail::decrement(impl_);
        return *this;
    }

    port_iterator port_iterator::operator--(int) {
        port_iterator it(*this);
        detail::decrement(impl_);
        return it;
    }

    //-----------------------------------------------------------

    ibv_mtu mtu_to_enum(unsigned mtu) {
        switch (mtu) {
            case 256:  return IBV_MTU_256;
            case 512:  return IBV_MTU_512;
            case 1024: return IBV_MTU_1024;
            case 2048: return IBV_MTU_2048;
            case 4096: return IBV_MTU_4096;
            default:   return static_cast<ibv_mtu>(-1);
        }
    }
}

