/*
 * Copyright (C) 2008 United States Government as represented by the
 * Administrator of the National Aeronautics and Space Administration.
 * All Rights Reserved.

Disclaimers

No Warranty: THE SUBJECT SOFTWARE IS PROVIDED 'AS IS' WITHOUT ANY WARRANTY
OF ANY KIND, EITHER EXPRESSED, IMPLIED, OR STATUTORY, INCLUDING, BUT NOT
LIMITED TO, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL CONFORM TO
SPECIFICATIONS, ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
PARTICULAR PURPOSE, OR FREEDOM FROM INFRINGEMENT, ANY WARRANTY THAT THE
SUBJECT SOFTWARE WILL BE ERROR FREE, OR ANY WARRANTY THAT DOCUMENTATION, IF
PROVIDED, WILL CONFORM TO THE SUBJECT SOFTWARE. THIS AGREEMENT DOES NOT, IN
ANY MANNER, CONSTITUTE AN ENDORSEMENT BY GOVERNMENT AGENCY OR ANY PRIOR
RECIPIENT OF ANY RESULTS, RESULTING DESIGNS, HARDWARE, SOFTWARE PRODUCTS OR
ANY OTHER APPLICATIONS RESULTING FROM USE OF THE SUBJECT SOFTWARE.
FURTHER, GOVERNMENT AGENCY DISCLAIMS ALL WARRANTIES AND LIABILITIES
REGARDING THIRD-PARTY SOFTWARE, IF PRESENT IN THE ORIGINAL SOFTWARE, AND
DISTRIBUTES IT 'AS IS.'

Waiver and Indemnity: RECIPIENT AGREES TO WAIVE ANY AND ALL CLAIMS AGAINST
THE UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS, AS WELL AS
ANY PRIOR RECIPIENT. IF RECIPIENT'S USE OF THE SUBJECT SOFTWARE RESULTS IN
ANY LIABILITIES, DEMANDS, DAMAGES, EXPENSES OR LOSSES ARISING FROM SUCH USE,
INCLUDING ANY DAMAGES FROM PRODUCTS BASED ON, OR RESULTING FROM, RECIPIENT'S
USE OF THE SUBJECT SOFTWARE, RECIPIENT SHALL INDEMNIFY AND HOLD HARMLESS THE
UNITED STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS, AS WELL AS ANY
PRIOR RECIPIENT, TO THE EXTENT PERMITTED BY LAW. RECIPIENT'S SOLE REMEDY
FOR ANY SUCH MATTER SHALL BE THE IMMEDIATE, UNILATERAL TERMINATION OF THIS
AGREEMENT.

 */

#ifndef _ibvpp_h_
#define _ibvpp_h_

#include <infiniband/verbs.h>

#include <tr1/memory>
#include <stdexcept>
#include <string>
#include <map>
#include <string.h>

#define IBVPP_SPNS std::tr1

namespace ibv {

    //! ibverbs API error
    class error : public std::runtime_error
    {
    public:
        int errcode; //<! The returned error code
        /*! \brief constructor
          * \param errstr A descriptive string.
          * \param err An error code.
          */
        explicit error(const std::string &errstr,int err=0)
            : std::runtime_error(errstr), errcode(err) {}
    };

    //-----------------------------------------------------------
    // forward declarations

    class port;

    //! ib port shared pointer
    typedef IBVPP_SPNS::shared_ptr<port> port_ptr;
    //! ib const port shared pointer
    typedef IBVPP_SPNS::shared_ptr<port const> const_port_ptr;

    class comp_channel;

    //! ib completion event channel shared pointer
    typedef IBVPP_SPNS::shared_ptr<comp_channel> comp_channel_ptr;
    //! ib const completion event channel shared pointer
    typedef IBVPP_SPNS::shared_ptr<comp_channel const> const_comp_channel_ptr;

    //! ib completion event channel weak pointer
    typedef IBVPP_SPNS::weak_ptr<comp_channel> comp_channel_weak_ptr;
    //! ib const completion event channel weak pointer
    typedef IBVPP_SPNS::weak_ptr<comp_channel const> const_comp_channel_weak_ptr;

    class cq;

    //! ib completion queue shared pointer
    typedef IBVPP_SPNS::shared_ptr<cq> cq_ptr;
    //! ib const completion queue channel shared pointer
    typedef IBVPP_SPNS::shared_ptr<cq const> const_cq_ptr;

    //! ib completion queue weak pointer
    typedef IBVPP_SPNS::weak_ptr<cq> cq_weak_ptr;
    //! ib const completion queue channel weak pointer
    typedef IBVPP_SPNS::weak_ptr<cq const> const_cq_weak_ptr;

    class mr;

    //! ib memory region shared pointer
    typedef IBVPP_SPNS::shared_ptr<mr> mr_ptr;
    //! ib const memory region shared pointer
    typedef IBVPP_SPNS::shared_ptr<mr const> const_mr_ptr;

    //! ib memory region weak pointer
    typedef IBVPP_SPNS::weak_ptr<mr> mr_weak_ptr;
    //! ib const memory region weak pointer
    typedef IBVPP_SPNS::weak_ptr<mr const> const_mr_weak_ptr;

    class pd;

    //! ib protection domain pointer
    typedef IBVPP_SPNS::shared_ptr<pd> pd_ptr;
    //! ib const protection domain pointer
    typedef IBVPP_SPNS::shared_ptr<pd const> const_pd_ptr;

    class srq;

    //! ib shared receive queue pointer
    typedef IBVPP_SPNS::shared_ptr<srq> srq_ptr;
    //! ib const shared receive queue pointer
    typedef IBVPP_SPNS::shared_ptr<srq const> const_srq_ptr;

    class ah;

    //! ib address handle pointer
    typedef IBVPP_SPNS::shared_ptr<ah> ah_ptr;
    //! ib const address handle pointer
    typedef IBVPP_SPNS::shared_ptr<ah const> const_ah_ptr;

    class qp;

    //! ib queue pair shared pointer
    typedef IBVPP_SPNS::shared_ptr<qp> qp_ptr;
    //! ib const queue pair shared pointer
    typedef IBVPP_SPNS::shared_ptr<qp const> const_qp_ptr;

    class device_list;

    //! ib device list shared pointer
    typedef IBVPP_SPNS::shared_ptr<device_list> device_list_ptr;

    //! get the infiniband device list
    device_list_ptr get_device_list() throw (std::exception);

    //-----------------------------------------------------------
    // attributes

    //! ib device attributes
    typedef ibv_device_attr device_attr;
    //! ib device attributes pointer
    typedef IBVPP_SPNS::shared_ptr<device_attr> device_attr_ptr;
    //! ib const device attributes pointer
    typedef IBVPP_SPNS::shared_ptr<device_attr const> const_device_attr_ptr;

    //! clear device attribute structure
    inline void clear(device_attr &attrs) {
        memset(&attrs,0,sizeof(device_attr));
    }

    //! ib port attributes
    typedef ibv_port_attr port_attr;
    //! ib port attributes pointer
    typedef IBVPP_SPNS::shared_ptr<port_attr> port_attr_ptr;
    //! ib const port attributes pointer
    typedef IBVPP_SPNS::shared_ptr<port_attr const> const_port_attr_ptr;

    //! clear port attribute structure
    inline void clear(port_attr &attrs) {
        memset(&attrs,0,sizeof(port_attr));
    }

    //! ib queue pair capabilities structure (part of qp init attributes)
    typedef ibv_qp_cap qp_cap;

    //! ib queue pair init attributes
    struct qp_init_attr {
        void            *qp_context; //!< Associated context of the QP
        cq_ptr           send_cq;    //!< CQ to be associated with the Send Queue (SQ)
        cq_ptr           recv_cq;    //!< CQ to be associated with the Receive Queue (RQ)
        srq_ptr          srq;        //!< SRQ handle if QP is to be associated with an SRQ, otherwise NULL
        qp_cap           cap;        //!< QP capabilities
        enum ibv_qp_type qp_type;    //!< QP Transport Service Type: IBV_QPT_RC, IBV_QPT_UC, or IBV_QPT_UD
        int              sq_sig_all; //!< If set, each Work Request (WR) submitted to the SQ generates a completion entry
        qp_init_attr() : sq_sig_all(0) {
            memset(&cap,0,sizeof(qp_cap));
        }
    };

    //! ib queue pair init attributes pointer
    typedef IBVPP_SPNS::shared_ptr<qp_init_attr> qp_init_attr_ptr;
    //! ib const queue pair init attributes pointer
    typedef IBVPP_SPNS::shared_ptr<qp_init_attr const> const_qp_init_attr_ptr;

    //! ib queue pair attribute mask
    typedef ibv_qp_attr_mask qp_attr_mask;

    //! ib queue pair attributes
    typedef ibv_qp_attr qp_attr;

    //! clear qp attribute structure
    inline void clear(qp_attr &attrs) {
        memset(&attrs,0,sizeof(qp_attr));
    }

    //! ib queue pair attributes pointer
    typedef IBVPP_SPNS::shared_ptr<qp_attr> qp_attr_ptr;
    //! ib const queue pair attributes pointer
    typedef IBVPP_SPNS::shared_ptr<qp_attr const> const_qp_attr_ptr;

    typedef ibv_srq_init_attr srq_init_attr;
    //! clear the shared receive queue init attributes structure
    inline void clear(srq_init_attr &attr) {
        memset(&attr,0,sizeof(attr));
    }

    //! ib shared receive queue attributes
    typedef ibv_srq_attr srq_attr;
    //! ib shared receive queue attribute mask
    typedef ibv_srq_attr_mask srq_attr_mask;

    //! ib address handle attributes
    typedef ibv_ah_attr ah_attr;
    //! clear the address handle attributes structure
    inline void clear(ah_attr &attr) {
        memset(&attr,0,sizeof(attr));
    }

    //-----------------------------------------------------------
    // structures

    //! receive-work-request structure
    typedef ibv_recv_wr recv_wr;
    //! clear receive work-request structure
    inline void clear(recv_wr &wr) {
        memset(&wr,0,sizeof(wr));
    }

    //! send-work-request structure
    typedef ibv_send_wr send_wr;
    //! clear send work-request structure
    inline void clear(send_wr &wr) {
        memset(&wr,0,sizeof(wr));
    }

    //! scatter/gather entry structure
    typedef ibv_sge sge;
    //! work completion structure
    typedef ibv_wc  wc;
    //! async event structure
    typedef ibv_async_event async_event;

    //-----------------------------------------------------------
    // ibv_context

    //! ib context
    class context : public IBVPP_SPNS::enable_shared_from_this<context> {
        ibv_context *context_;
        mutable device_attr_ptr device_attrs;

    public:

        context(ibv_context *context_in) : context_(context_in) {}
        ~context();

        //! Get the ibv_context.
        ibv_context *get() { return context_; }
        //! Get the ibv_context.
        const ibv_context *get() const { return context_; }

        //! Get device properties
        void query_device(device_attr *device_attr) const
            throw (error);

        //! Get device properties
        const_device_attr_ptr get_device_attrs() const
            throw (std::exception);

        //!  Get port properties
        void query_port(uint8_t port_num, port_attr *port_attr) const
            throw (error);

        //! Get a GID table entry
        void query_gid(uint8_t port_num, int index, ibv_gid *gid) const
            throw (error);

        //! Get a P_Key table entry
        void query_pkey(uint8_t port_num, int index, uint16_t *pkey) const
            throw (error);

        //! get a device's name
        const char *get_device_name() const {
            return ibv_get_device_name(context_->device);
        }

        //! get a port object
        port_ptr get_port(uint8_t port_num);
        //! get a const port object
        const_port_ptr get_port(uint8_t port_num) const;

        //! Allocate a protection domain.
        inline pd_ptr alloc_pd();

        //! Create a completion queue
        cq_ptr create_cq(int cqe,void *cq_context,
                         const comp_channel_ptr &cchan=comp_channel_ptr(),
                         int comp_vector=0) throw (std::exception);

        //! Create a completion event channel
        inline comp_channel_ptr create_comp_channel();

        //! Get next async event
        void get_async_event(async_event *event) {
            if (ibv_get_async_event(context_,event) != 0)
                throw error("get_async_event() failed");
        }
    };

    //! ib context shared pointer
    typedef IBVPP_SPNS::shared_ptr<context> context_ptr;
    //! ib const context shared pointer
    typedef IBVPP_SPNS::shared_ptr<context const> const_context_ptr;
 
    //! Acknowledge an async event
    inline void ack_async_event(async_event *event) {
        ibv_ack_async_event(event);
    }

    //-----------------------------------------------------------
    // port

    class port {
        context_ptr context_;
        uint8_t portnum_;

    public:
        port(const context_ptr &ctx,uint8_t pnum)
            : context_(ctx), portnum_(pnum) {}
        ~port() {}

        //! get the ib device context pointer
        const context_ptr &get_context() { return context_; }
        //! get the ib device context pointer
        const_context_ptr get_context() const { return context_; }

        //! get the port's number
        uint8_t get_num() const { return portnum_; }

        //! get a string that uniquely names this port ('devicename'.'portnum')
        std::string get_name() const;

        //! get the port's LID
        uint16_t get_lid() const;

        //!  Get port properties
        void query(port_attr *port_attr) const
        {
            context_->query_port(portnum_,port_attr);
        }

        //! Get a P_Key table entry
        void query_pkey(int index, uint16_t *pkey) const
        {
            context_->query_pkey(portnum_,index,pkey);
        }
    };

    //-----------------------------------------------------------
    // ibv_pd

    //! ib protection domain
    class pd  : public IBVPP_SPNS::enable_shared_from_this<pd> {
        ibv_pd *pd_;
        context_ptr context_;

    public:
        pd(const context_ptr &ctx) throw (std::exception);
        ~pd();

        //! Get the ibv_pd.
        ibv_pd *get() { return pd_; }
        //! Get the ibv_pd.
        const ibv_pd *get() const { return pd_; }

        //! get the ib device context pointer
        const context_ptr &get_context() { return context_; }
        //! get the ib device context pointer
        const_context_ptr get_context() const { return context_; }

        //! Register a memory region.
        mr_ptr reg_mr(void *addr,size_t length,ibv_access_flags access);

        //! Create a queue pair.
        inline qp_ptr create_qp(qp_init_attr *qp_init_attr);

        //! Creates a SRQ associated with the specified protection domain.
        inline srq_ptr create_srq(srq_init_attr *srq_init_attr);

        //! Create an address handle.
        inline ah_ptr create_ah(ah_attr *ah_attr);
    };

    //! Allocate a protection domain.
    inline pd_ptr context::alloc_pd() {
        return pd_ptr(new pd(shared_from_this()));
    }

    //-----------------------------------------------------------
    // ibv_mr

    //! ib memory region
    class mr {
        ibv_mr *mr_;
        pd_ptr pd_;
    public:
        mr(const pd_ptr &pdp,void *addr,size_t length,ibv_access_flags access)
            throw (std::exception);
        ~mr();

        //! Get the ibv_mr.
        ibv_mr *get() { return mr_; }
        //! Get the ibv_mr.
        const ibv_mr *get() const { return mr_; }

        //! get the memory region's lkey
        uint32_t get_lkey() const { return mr_->lkey; }
        //! get the memory region's rkey
        uint32_t get_rkey() const { return mr_->rkey; }

        //! get the ib protection domain pointer
        const pd_ptr &get_pd() { return pd_; }
        //! get the ib protection domain pointer
        const_pd_ptr get_pd() const { return pd_; }
    };

    //-----------------------------------------------------------
    // ibv_srq

    //! ib shared receive queue
    class srq {
        ibv_srq *srq_;
        pd_ptr pd_;
    public:
        srq(const pd_ptr &pdp,srq_init_attr *init_attr)
            throw (error);
        ~srq();

        //! Get the ibv_srq.
        ibv_srq *get() { return srq_; }
        //! Get the ibv_srq.
        const ibv_srq *get() const { return srq_; }

        //! get the ib protection domain pointer
        const pd_ptr &get_pd() { return pd_; }
        //! get the ib protection domain pointer
        const_pd_ptr get_pd() const { return pd_; }

        //! Modifies the attributes for the specified SRQ.
        void modify(srq_attr *attr,srq_attr_mask attr_mask) throw (error);

        //! Returns the attribute list and current values for the specified SRQ.
        void query(srq_attr *attr) const throw (error);

        //! Posts a list of work requests to the specified SRQ.
        int post_recv(recv_wr *wr,recv_wr **bad_wr) {
            return ibv_post_srq_recv(srq_,wr,bad_wr);
        }
        //! Posts a list of work requests to the specified SRQ.
        void post_recv(recv_wr *wr) throw (error)
        {
            recv_wr *bad_wr;
            int result;
            if ((result=post_recv(wr,&bad_wr)))
                throw error("ibv_post_srq_recv", result);
        }
    };

    inline srq_ptr pd::create_srq(srq_init_attr *srq_init_attr)
    {
        return srq_ptr(new srq(shared_from_this(),srq_init_attr));
    }

    //-----------------------------------------------------------
    // ibv_ah

    //! ib address handle
    class ah {
        ibv_ah *ah_;
        pd_ptr pd_;
    public:
        ah(const pd_ptr &pdp,ah_attr *attr) throw (error);
        ~ah();

        //! Get the ibv_ah.
        ibv_ah *get() { return ah_; }
        //! Get the ibv_ah.
        const ibv_ah *get() const { return ah_; }

        //! get the ib protection domain pointer
        const pd_ptr &get_pd() { return pd_; }
        //! get the ib protection domain pointer
        const_pd_ptr get_pd() const { return pd_; }
    };

    inline ah_ptr pd::create_ah(ah_attr *ah_attr)
    {
        return ah_ptr(new ah(shared_from_this(),ah_attr));
    }

    //-----------------------------------------------------------
    // ibv_qp

    //! ib queue pair
    class qp {
        ibv_qp *qp_;
        pd_ptr pd_;
        cq_ptr scq_, rcq_;
        srq_ptr srq_;
    public:
        qp(const pd_ptr &pdp,qp_init_attr *init_attr)
            throw (error);
        ~qp();

        //! Get the ibv_qp.
        ibv_qp *get() { return qp_; }
        //! Get the ibv_qp.
        const ibv_qp *get() const { return qp_; }

        //! get the ib protection domain pointer
        const pd_ptr &get_pd() { return pd_; }
        //! get the ib protection domain pointer
        const_pd_ptr get_pd() const { return pd_; }

        //! get the receive completion queue pointer
        const cq_ptr &get_rcq() { return rcq_; }
        //! get the receive completion queue pointer
        const_cq_ptr get_rcq() const { return rcq_; }

        //! get the send completion queue pointer
        const cq_ptr &get_scq() { return scq_; }
        //! get the send completion queue pointer
        const_cq_ptr get_scq() const { return scq_; }

        //! get the shared receive queue pointer
        const srq_ptr &get_srq() { return srq_; }
        //! get the shared receive queue pointer
        const_srq_ptr get_srq() const { return srq_; }

        //! Get the ibv_qp's number.
        uint32_t get_num() const { return qp_->qp_num; }

        //! Modify a queue pair.
        void modify(qp_attr *attr,qp_attr_mask attr_mask) throw (error);

        //! Returns the attribute list and current values for the specified QP
        void query(qp_attr *attr,
                  qp_attr_mask attr_mask,
                  qp_init_attr *init_attr) const throw (error);

        //! Post a list of work requests to a send queue.
        int post_send(send_wr *wr,send_wr **bad_wr) {
            return ibv_post_send(qp_,wr,bad_wr);
        }
        void post_send(send_wr *wr) throw (error) {
            send_wr *bad_wr;
            int result;
            if ((result=post_send(wr,&bad_wr)))
                throw error("ibv_post_send", result);
        }

        //! Post a list of work requests to a receive queue.
        int post_recv(recv_wr *wr,recv_wr **bad_wr) {
            return ibv_post_recv(qp_,wr,bad_wr);
        }
        //! Post a list of work requests to a receive queue.
        void post_recv(recv_wr *wr) throw (error)
        {
            recv_wr *bad_wr;
            int result;
            if ((result=post_recv(wr,&bad_wr)))
                throw error("ibv_post_recv", result);
        }

        void attach_mcast(ibv_gid *gid,uint16_t lid)
        {
            int result;
            if ((result=ibv_attach_mcast(qp_,gid,lid)))
                throw error("ibv_attach_mcast", result);
        }

        void detach_mcast(ibv_gid *gid,uint16_t lid)
        {
            int result;
            if ((result=ibv_detach_mcast(qp_,gid,lid)))
                throw error("ibv_detach_mcast", result);
        }
    };

    inline qp_ptr pd::create_qp(qp_init_attr *qp_init_attr)
    {
        return qp_ptr(new qp(shared_from_this(),qp_init_attr));
    }

    //-----------------------------------------------------------
    // ibv_cq

    //! ib completion queue
    class cq {
        friend cq_ptr context::create_cq(int cqe,
                                         void *cq_context,
                                         const comp_channel_ptr &cchan,
                                         int comp_vector)
            throw (std::exception);

        ibv_cq *cq_;
        context_ptr context_;
        IBVPP_SPNS::weak_ptr<comp_channel> comp_channel_;

        cq(const context_ptr &ctx,ibv_cq *cq) : cq_(cq), context_(ctx) {}

    public:
        ~cq();

        //! Get the ibv_cq.
        ibv_cq *get() { return cq_; }
        //! Get the ibv_cq.
        const ibv_cq *get() const { return cq_; }

        //! get the ib device context pointer
        const context_ptr &get_context() { return context_; }
        //! get the ib device context pointer
        const_context_ptr get_context() const { return context_; }

        //! Modify the capacity of the CQ.
        void resize(int cqe) throw (error);

        //! Acknowledge CQ completion events.
        void ack_events(unsigned int nevents) throw ();

        //! Poll a CQ for work completions
        unsigned poll(int num_entries,wc *wc) throw (error);

        //! Request completion notification on a CQ.
        void req_notify(bool solicited_only) throw (error);
    };

    //-----------------------------------------------------------
    // ibv_comp_channel

    //! ib completion event channel
    class comp_channel {
        friend cq_ptr context::create_cq(int cqe,
                                         void *cq_context,
                                         const comp_channel_ptr &cchan,
                                         int comp_vector)
            throw (std::exception);
        friend class cq;

        ibv_comp_channel *comp_channel_;
        context_ptr context_;
        typedef std::map<uint32_t,IBVPP_SPNS::weak_ptr<cq> > cq_map_type;
        cq_map_type cq_map;

    public:
        comp_channel(const context_ptr &ctx);
        ~comp_channel();

        //! Get the ibv_comp_channel.
        ibv_comp_channel *get() { return comp_channel_; }
        //! Get the ibv_comp_channel.
        const ibv_comp_channel *get() const { return comp_channel_; }

        //! get the ib device context pointer
        const context_ptr &get_context() { return context_; }
        //! get the ib device context pointer
        const_context_ptr get_context() const { return context_; }

        //! Read next CQ event.
        cq_ptr get_event(void **cq_context) throw (error);
    };

    inline comp_channel_ptr context::create_comp_channel()
    {
        return comp_channel_ptr(new comp_channel(shared_from_this()));
    }

    //-----------------------------------------------------------
    // ibv_device

    class device {
        friend class device_list;
        ibv_device *device_;
        device_list_ptr devlist_;
        device(ibv_device *d,const device_list_ptr &l)
            : device_(d), devlist_(l) {}

    public:
        //! Get the ibv_device.
        ibv_device *get() { return device_; }
        //! Get the ibv_device.
        const ibv_device *get() const { return device_; }

        //! get a device's name
        const char *get_device_name() {
            return ibv_get_device_name(device_);
        }
        //! get a device's GUID
        uint64_t get_device_guid() {
            return ibv_get_device_guid(device_);
        }

        //! open a device
        context_ptr open_device() throw (std::exception);
    };

    //! ib device shared pointer
    typedef IBVPP_SPNS::shared_ptr<device> device_ptr;

    //-----------------------------------------------------------
    // ibv_device_list

    class port_iterator;

    //! ib device list
    class device_list : public IBVPP_SPNS::enable_shared_from_this<device_list>
    {
        friend device_list_ptr get_device_list() throw (std::exception);

        ibv_device **device_list_;
        size_t device_count_;

        device_list(ibv_device **device_list_in,size_t device_count_in)
            : device_list_(device_list_in), device_count_(device_count_in)
        {}

    public:
        ~device_list() throw ();

        //! ib device list iterator (random access)
        class iterator {
            ibv_device **device_;
            friend class device_list;
            iterator(ibv_device **l) : device_(l) {}
        public:
            iterator() {}
            iterator(const iterator &it) : device_(it.device_) {}
            iterator &operator=(const iterator &it) {
                device_ = it.device_;
                return *this;
            }
            iterator &operator++() {
                ++device_;
                return *this;
            }
            iterator operator++(int) {
                iterator it(*this);
                device_++;
                return it;
            }
            iterator &operator--() {
                --device_;
                return *this;
            }
            iterator operator--(int) {
                iterator it(*this);
                device_--;
                return it;
            }

            template <class T>
            friend iterator operator+(const iterator &lhs,T rhs) {
                return iterator(lhs.device_ + rhs);
            }
            template <class T>
            friend iterator operator-(const iterator &lhs,T rhs) {
                return iterator(lhs.device_ - rhs);
            }

            friend bool operator==(const iterator &lhs,const iterator &rhs) {
                return lhs.device_ == rhs.device_;
            }
            friend bool operator!=(const iterator &lhs,const iterator &rhs) {
                return lhs.device_ != rhs.device_;
            }

            //! get a pointer to the pointed-to ibv_device
            ibv_device *operator*() const {
                if (*device_ == 0)
                    throw error("end device iterator");
                return *device_;
            }
        };

        //! get the iterator at the beginning of the device list.
        iterator begin() { return iterator(device_list_); }
        //! get the iterator at the end of the device list.
        iterator end() { return iterator(&device_list_[device_count_]); }

        //! test whether there are any devices
        bool empty() const { return (device_count_ == 0); }

        //! get the number of devices
        size_t size() const { return device_count_; }

        //! get the port_iterator at the beginning of the port list.
        inline port_iterator begin_port();
        //! get the port_iterator at the end of the port list.
        inline port_iterator end_port();

        //! get a device's name
        const char *get_device_name(const iterator &it) {
            return ibv_get_device_name(*it);
        }
        //! get a device's GUID
        uint64_t get_device_guid(const iterator &it) {
            return ibv_get_device_guid(*it);
        }

        //! get a counted reference to a device
        device_ptr get_device(const iterator &it)
        {
            return device_ptr(new device(*it,shared_from_this()));
        }

        //! open a device
        context_ptr open_device(const iterator &it) throw (std::exception);
    };

    //-----------------------------------------------------------
    // port iterator

    class port_iterator {
    public:
        typedef port_iterator type;

        typedef std::pair<device_list::iterator,uint8_t> position_type;

        port_iterator(const device_list::iterator &dit,uint8_t pnum);

        port_iterator() {}
        port_iterator(const port_iterator &it) : impl_(it.impl_) {}
        port_iterator &operator=(const port_iterator &it) {
            impl_ = it.impl_;
            return *this;
        }
        port_iterator &operator++();
        port_iterator operator++(int);
        port_iterator &operator--();
        port_iterator operator--(int);

        friend bool operator==(const type &lhs,const type &rhs) {
            return lhs.impl_.pos_ == rhs.impl_.pos_;
        }
        friend bool operator!=(const type &lhs,const type &rhs) {
            return lhs.impl_.pos_ != rhs.impl_.pos_;
        }

        //! get the pointed-to device/portnum pair
        port_ptr operator*() const;

    private:
        struct Impl_ {
            position_type pos_;
            mutable context_ptr context_;
            mutable port_ptr port_;
            Impl_() {}
            Impl_(const device_list::iterator &dit,uint8_t pnum)
                : pos_(dit,pnum) {}
        };
        mutable Impl_ impl_;

        class detail;
    };

    port_iterator device_list::begin_port() {
        return port_iterator(begin(),0);
    }

    port_iterator device_list::end_port() {
        return port_iterator(end(),0);
    }

    //-----------------------------------------------------------

    //! convert an mtu value to the appropriate enum
    ibv_mtu mtu_to_enum(unsigned mtu);
}

#endif

